package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;
import at.ac.uibk.gitsearch.properties.ApplicationProperties;
import at.ac.uibk.gitsearch.service.dto.MetadataUserDTO;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.json.JsonMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.dataformat.yaml.YAMLGenerator;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.zip.ZipInputStream;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import net.minidev.json.JSONObject;
import net.minidev.json.JSONStyle;
import org.apache.commons.io.FileUtils;
import org.eclipse.jgit.lib.Repository;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.Group;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

@Service
@SuppressWarnings({ "PMD.GodClass", "PMD.CognitiveComplexity", "PMD.CyclomaticComplexity", "PMD.NPathComplexity" })
public class ExerciseImportService {

    private static final String DIFFICULTY = "difficulty";

    private static final String TMP_DIR_PROPERTY = "java.io.tmpdir";

    private final Logger log = LoggerFactory.getLogger(ExerciseImportService.class);

    private final ApplicationProperties applicationProperties;

    private final GitlabService gitlabService;

    private final ZipService zipService;

    public ExerciseImportService(ApplicationProperties applicationProperties, ZipService zipService, GitlabService gitlabService) {
        this.applicationProperties = applicationProperties;
        this.zipService = zipService;
        this.gitlabService = gitlabService;
    }

    /**
     * Retrieve a zipped Artemis Exercise from a given URL
     * and store it in a temporary directory
     *
     * @param exerciseUrl the url of the exposed exercise
     * @throws ArtemisImportError
     */
    @SuppressWarnings("PMD.AvoidCatchingGenericException")
    public void importExerciseFromArtemis(String exerciseUrl) throws ArtemisImportError {
        String exerciseToken = getTokenFromUrl(exerciseUrl);

        try (ZipInputStream zipInput = getZipInputStreamFromUrl(exerciseUrl);) {
            zipService.storeZipInTmp(zipInput, exerciseToken);
        } catch (URISyntaxException e) {
            log.error("Unable to add apiKey to the import request for [{}]. Try to request resource without apiKey.", exerciseUrl);
        } catch (Exception e) {
            throw new ArtemisImportError(
                "ZipFile couldn't be retrieved from [" + URLDecoder.decode(exerciseUrl, StandardCharsets.UTF_8) + "]" + e.getMessage(),
                e
            );
        }
    }

    protected ZipInputStream getZipInputStreamFromUrl(String exerciseUrl) throws URISyntaxException {
        Client client = ClientBuilder.newClient();
        WebTarget target = client.target(exerciseUrl);

        InputStream inputStream = target
            .request()
            .header("Authorization", getApiKeyOfRegisteredConnectorByUrl(new URI(exerciseUrl).getAuthority()))
            .accept(javax.ws.rs.core.MediaType.APPLICATION_OCTET_STREAM)
            .get(InputStream.class);
        return new ZipInputStream(inputStream);
    }

    /**
     * Import exercise to Gitlab.
     * Sequentially executes following processes:
     * - Apply user defined exercise metadata
     * - Create repository in Gitlab's group defined by the user
     * - Locally checking out the new created repository
     * - Copy, commit and push files to the new repository
     *
     * @param exerciseInfo  of the exercise to import
     * @param exerciseToken of the stored exercise (in the temporary directory) to
     *                      import
     */
    public void importExercise(ArtemisExerciseInfo exerciseInfo, String exerciseToken, Long gitlabGroupId)
        throws GitLabApiException, IOException, ArtemisImportError {
        sanitizeExerciseInfo(exerciseInfo);
        validateExerciseInfo(exerciseInfo);

        applyExerciseInfoChanges(exerciseInfo, exerciseToken);
        renameFiles(exerciseToken);

        gitlabService.setJGitDefaultAuth();
        Group group = gitlabService.getGroupById(gitlabGroupId);
        String repoUrl = gitlabService.createRepository(group, exerciseInfo.getTitle());
        try (Repository repo = gitlabService.checkoutRepo(repoUrl)) {
            gitlabService.commitAndPushToRepo(repo, new File(System.getProperty(TMP_DIR_PROPERTY), exerciseToken));
            FileUtils.deleteDirectory(new File(System.getProperty(TMP_DIR_PROPERTY), exerciseToken));
            FileUtils.deleteDirectory(repo.getDirectory().getParentFile());
        } catch (IOException e) {
            log.error("Could not remove directory: {} \n", e.getMessage());
        }
    }

    /**
     * Retrieves metadata from imported Artemis exercise
     *
     * @param exerciseToken of the imported exercise
     * @return Content of exercise's 'metadata.yaml' and 'artemis.yaml'
     */
    public ArtemisExerciseInfo getArtemisExerciseInfo(String exerciseToken) throws IOException {
        ObjectMapper mapper = JsonMapper
            .builder()
            .configure(DeserializationFeature.READ_ENUMS_USING_TO_STRING, true)
            .configure(MapperFeature.ACCEPT_CASE_INSENSITIVE_ENUMS, true)
            .configure(DeserializationFeature.ACCEPT_SINGLE_VALUE_AS_ARRAY, true)
            .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false)
            .build();

        File exerciseDetailsFile = Objects.requireNonNull(
            new File(System.getProperty(TMP_DIR_PROPERTY), exerciseToken)
                .listFiles((file, name) -> name.startsWith("Exercise-Details-") && name.endsWith(".json"))
        )[0];
        mapArtemisExerciseInfoValues(exerciseDetailsFile);

        ArtemisExerciseInfo artemisExerciseInfo = mapper.readValue(exerciseDetailsFile, ArtemisExerciseInfo.class);

        writeArtemisMetadataFile(exerciseDetailsFile, artemisExerciseInfo);
        return artemisExerciseInfo;
    }

    /**
     * Utility method used to map Artemis Exercise metadata values to the ones used
     * in
     * Sharing Metadata
     *
     * @param jsonFile to parse the values from
     * @throws IOException if an error occurs
     */
    private void mapArtemisExerciseInfoValues(File jsonFile) throws IOException {
        ObjectMapper mapper = new ObjectMapper();
        JSONObject jsonObject = mapper.readValue(jsonFile, JSONObject.class);
        if (jsonObject.containsKey(DIFFICULTY)) {
            try {
                jsonObject.put(DIFFICULTY, ArtemisExerciseInfo.DifficultyMapping.valueOf(jsonObject.get(DIFFICULTY).toString()));
            } catch (IllegalArgumentException iae) {
                log.warn(
                    "Artemis difficulty value '{}' could not be mapped to its equivalent Gitsearch value. Skipping..",
                    jsonObject.get(DIFFICULTY)
                );
                jsonObject.remove(DIFFICULTY);
            }
        }

        if (!jsonObject.containsKey("metadataVersion")) {
            jsonObject.put("metadataVersion", "0.4");
        }

        if (!jsonObject.containsKey("license")) {
            jsonObject.put("license", "CC-SA-BY 4.0");
        }

        if (!jsonObject.containsKey("format")) {
            jsonObject.put("format", "artemis");
        }

        if (!jsonObject.containsKey("learningResourceType")) {
            jsonObject.put("learningResourceType", "artemis");
        }

        if (!jsonObject.containsKey("structure")) {
            jsonObject.put("structure", "atomic");
        }

        if (jsonObject.containsKey("instructors")) {
            jsonObject.put("creator", jsonObject.get("instructors"));
            jsonObject.remove("instructors");
        }

        if ("programming".equals(jsonObject.get("type").toString())) {
            jsonObject.put("type", "programming exercise");
        }

        try (FileWriter fileWriter = new FileWriter(jsonFile, StandardCharsets.UTF_8)) {
            fileWriter.write(jsonObject.toString(JSONStyle.NO_COMPRESS));
        }
    }

    /**
     * Utility method used to extract Artemis metadata into a 'artemis.yaml' file
     *
     * @param jsonMetadataFile    where the exercise metadata is stored
     * @param artemisExerciseInfo the entity of the stored metadata
     * @throws IOException if an error occurs
     */
    private void writeArtemisMetadataFile(File jsonMetadataFile, ArtemisExerciseInfo artemisExerciseInfo) throws IOException {
        ObjectMapper mapper = new ObjectMapper(
            new YAMLFactory().enable(YAMLGenerator.Feature.INDENT_ARRAYS).disable(YAMLGenerator.Feature.WRITE_DOC_START_MARKER)
        );

        File artemisMetadataFile = new File(jsonMetadataFile.getParent(), "artemis.yaml");
        if (!artemisMetadataFile.exists()) {
            boolean success = artemisMetadataFile.createNewFile();
            if (!success) {
                log.warn("Cannot create file");
            }
        }
        mapper.writeValue(artemisMetadataFile, artemisExerciseInfo.getArtemisInfoOnly());
    }

    /**
     * Utility method used to apply metadata changes to the exercise's
     * 'metadata.yaml' file
     *
     * @param exerciseInfo  containing the metadata values
     * @param exerciseToken of the exercise to update
     * @throws IOException
     */
    public void applyExerciseInfoChanges(ArtemisExerciseInfo exerciseInfo, String exerciseToken) throws IOException {
        ObjectMapper mapper = new ObjectMapper(
            new YAMLFactory().enable(YAMLGenerator.Feature.INDENT_ARRAYS).disable(YAMLGenerator.Feature.WRITE_DOC_START_MARKER)
        );
        mapper.writeValue(
            Paths.get(System.getProperty(TMP_DIR_PROPERTY), exerciseToken, "metadata.yaml").toFile(),
            exerciseInfo.getExerciseInfoOnly()
        );
    }

    /**
     * Used to validate apiKeys by checking if they match with
     * one of the registeredConnectors' keys stored in application-(dev|prod).yml
     *
     * @param apiKey the apikey to validate
     * @return true if equal to the local one, false otherwise
     */
    public boolean validate(String apiKey) {
        for (ApplicationProperties.RegisteredConnector registeredConnector : applicationProperties.getRegisteredConnectors()) {
            if (registeredConnector.getAccessToken().equals(apiKey)) {
                log.debug("ApiKey validated correctly against [{} ]", registeredConnector.getUrl());
                return true;
            }
        }
        return false;
    }

    /**
     * Method that retrieves the ApiKey of a Registered Connector configured in
     * 'application-dev/prod.yml'
     * given the URL of the connector
     *
     * @param urlAuthority URL Authority/Domain of the connector
     * @return the ApiKey, or Null if no connector with the given url is found
     */
    private String getApiKeyOfRegisteredConnectorByUrl(String urlAuthority) {
        for (ApplicationProperties.RegisteredConnector registeredConnector : applicationProperties.getRegisteredConnectors()) {
            if (registeredConnector.getUrl().contains(urlAuthority)) {
                return registeredConnector.getAccessToken();
            }
        }
        return null;
    }

    /**
     * Renames specific files of an imported Artemis exercise,
     * to meet naming conventions
     *
     * Renamed files:
     * - Problem-Statement-xxx.md -> exercise.md
     * - Exercise-Details-xxx.json -> Exercise-Details.json
     *
     * @param exerciseToken to retrieve the location of the files to rename
     */
    private void renameFiles(String exerciseToken) {
        try {
            File problemStatementFile = Objects.requireNonNull(
                new File(System.getProperty(TMP_DIR_PROPERTY), exerciseToken)
                    .listFiles((file, name) -> name.startsWith("Problem-Statement-") && name.endsWith(".md"))
            )[0];
            if (!problemStatementFile.renameTo(Paths.get(System.getProperty(TMP_DIR_PROPERTY), exerciseToken, "exercise.md").toFile())) {
                log.error("Unable to rename file [{}]. Skipping..", problemStatementFile.getName());
            }
            File exerciseDetailsFile = Objects.requireNonNull(
                new File(System.getProperty(TMP_DIR_PROPERTY), exerciseToken)
                    .listFiles((file, name) -> name.startsWith("Exercise-Details-") && name.endsWith(".json"))
            )[0];
            Files.move(exerciseDetailsFile.toPath(), exerciseDetailsFile.toPath().resolveSibling("Exercise-Details.json"));
        } catch (ArrayIndexOutOfBoundsException e) {
            log.error("File to rename could not be found in [{}].", System.getProperty(TMP_DIR_PROPERTY) + File.separator + exerciseToken);
        } catch (IOException ioe) {
            log.error("Error while renaming file: {}", ioe.getMessage());
        }
    }

    /**
     * Sanitize the provided input fields for the exercise information, by escaping
     * all potential tags' brackets
     *
     * @param exerciseInfo of the imported exercise
     */
    private void sanitizeExerciseInfo(ArtemisExerciseInfo exerciseInfo) {
        exerciseInfo.setMetadataVersion(exerciseInfo.getMetadataVersion());
        exerciseInfo.setTitle(replaceTagCharacters(exerciseInfo.getTitle()));
        exerciseInfo.setDescription(replaceTagCharacters(exerciseInfo.getDescription()));
        if (exerciseInfo.getKeyword() != null) {
            exerciseInfo.setKeyword(exerciseInfo.getKeyword().stream().map(this::replaceTagCharacters).collect(Collectors.toList()));
        }
        if (exerciseInfo.getCreator() != null) {
            exerciseInfo.setCreator(exerciseInfo.getCreator().stream().map(creator -> creator.sanitize()).collect(Collectors.toList()));
        }
        if (exerciseInfo.getPublisher() != null) {
            exerciseInfo.setPublisher(
                exerciseInfo.getPublisher().stream().map(publisher -> publisher.sanitize()).collect(Collectors.toList())
            );
        }
        exerciseInfo.setLicense(replaceTagCharacters(exerciseInfo.getLicense()));
        if (exerciseInfo.getLanguage() != null) {
            exerciseInfo.setLanguage(exerciseInfo.getLanguage().stream().map(this::replaceTagCharacters).collect(Collectors.toList()));
        }
        if (exerciseInfo.getFormat() != null) {
            exerciseInfo.setFormat(exerciseInfo.getFormat().stream().map(this::replaceTagCharacters).collect(Collectors.toList()));
        }
        exerciseInfo.setLearningResourceType(replaceTagCharacters(exerciseInfo.getLearningResourceType()));
    }

    /**
     * Method used to validate the exercise information of the imported exercise
     *
     * @param exerciseInfo of the imported exercise
     * @throws ArtemisImportError if the exercise's information is invalid or
     *                            incomplete
     */
    public void validateExerciseInfo(ArtemisExerciseInfo exerciseInfo) throws ArtemisImportError {
        if (!exerciseInfo.validateMandatoryFields()) { // Check mandatory fields
            throw new ArtemisImportError("exercise.import.error.invalidMetadata");
        }
        for (String keyword : exerciseInfo.getKeyword()) {
            if (keyword.isBlank()) {
                throw new ArtemisImportError("exercise.import.error.invalidKeyword");
            }
        }
        for (String programmingLanguage : exerciseInfo.getProgrammingLanguage()) {
            if (programmingLanguage.isBlank()) {
                throw new ArtemisImportError("exercise.import.error.invalidProgrammingLanguage");
            }
        }
        for (MetadataUserDTO creator : exerciseInfo.getCreator()) {
            if (!creator.isValid()) {
                throw new ArtemisImportError("exercise.import.error.invalidCreator");
            }
        }
        for (MetadataUserDTO publisher : exerciseInfo.getPublisher()) {
            if (!publisher.isValid()) {
                throw new ArtemisImportError("exercise.import.error.invalidPublisher");
            }
        }
    }

    /**
     * Method responsible for extracting an exercise's token from
     * its Url.
     * @param exerciseUrl to extract the token from
     * @return The exercise's token
     */
    public String getTokenFromUrl(String exerciseUrl) {
        return exerciseUrl.split("/")[exerciseUrl.split("/").length - 1];
    }

    private String replaceTagCharacters(String input) {
        if (input != null) {
            return input.replace("<", "&lt").replace(">", "&gt");
        }
        return null;
    }
}
